/*=============================================================================
	UnActor.h: Actor-related functions

	Copyright 1996 Epic MegaGames, Inc. This software is a trade secret.
	Compiled with Visual C++ 4.0. Best viewed with Tabs=4.

	Revision history:
		* Created by Tim Sweeney
        * Aug 30, 1996: Mark added PLevel
		* Oct 19, 1996: Tim redesigned to eliminate redundency
=============================================================================*/

#ifndef _INC_UNACTOR
#define _INC_UNACTOR

#include "UnMsgs.h"		/* Predefined actor messages		*/
#include "UnArray.h"	/* Actor array optimization code	*/

/*-----------------------------------------------------------------------------
	Forward definitions.
-----------------------------------------------------------------------------*/

class AActor;

/*-----------------------------------------------------------------------------
	Includes.
-----------------------------------------------------------------------------*/

//
// Include the root class definition generated by Unreal.
// This defines AActorBase, AActor, and all child helpers and child classes.
//
#ifndef IMPLEMENTATION_API
	#ifdef _DEBUG
		#define IMPLEMENTATION_API DLL_EXPORT /* DLL exporting of AActorBase-derived classes */
	#else
		#define IMPLEMENTATION_API UNGAME_API /* DLL exporting of AActorBase-derived classes */
	#endif
#endif

#include "Root.h"

#ifndef _INC_UNACTLIST		// Actor list resource class definition.
#include "UnActLst.h"
#endif

#ifndef _INC_UNCLSPRP		// Actor class property definitions.
#include "UnClsPrp.h"
#endif

#ifndef _INC_UNCLASS		// Actor class resource class definition.
#include "UnClass.h"
#endif

#ifndef _INC_UNSCRIPT		// Script resource class.
#include "UnScript.h"
#endif

#ifndef _INC_UNMSGPAR		// Actor message parameter classes.
#include "UnMsgPar.h"
#endif

/*-----------------------------------------------------------------------------
	Class-related global variables.
-----------------------------------------------------------------------------*/

//
// Actor types that are built into the engine.
// This structure is temporary and will become obsolete.
//
class UNREAL_API FGlobalClasses
{
public:

	// Resources.
	TArray<UClass>	*IntrinsicClasses;	// List of built-in, always-active classes.

	// Built-in actor classes: All of these classes are built into the engine and
	// are always loaded.  They are referenced in the IntrinsicClasses array, which
	// is referenced by to Root resource, which prevents them from being freed during
	// a resource purged.  Do not add additional classes here unless absolutely
	// necessary.  If you do add a class, make sure it's referenced by IntrinsicClasses
	// or else it will be purged and its pointer will be invalid.
	UClass			*Actor,*Camera,*Light,*LevelDescriptor;

	// Functions.
	void Init();
	void Exit();
	void Associate();
};

/*-----------------------------------------------------------------------------
	Global actor and class functions
-----------------------------------------------------------------------------*/

//
// Actor import/export functions (UnActLst.cpp).
//
void ExportActorProperty(char *Type,char *Name, char *Value, int iElement,
	AActor *Actor, int Flags, int Descriptive, int ArrayElement, AActor *Delta, FName Category);
int ExportActor(AActor *Actor,char *Ptr,FName PropertyName,
	int Indent,int Descriptive,int Flags, AActor *Delta, int Resources, int ArrayElement,FName Name);
int ExportMultipleActors (UActorList *Actors,char *Ptr,FName Name,
	int Indent,int Descriptive,FName Category);
const char *ImportActorProperties(AActor *Actor, const char *Data);
int PropertyFrameEndOffset(int NumFrameProperties,const FClassProperty *FrameProperties,int NewElementSize);

//
// Script compiler functions (UnScript.cpp).
//
int UNREAL_API MakeScripts(UClass *Class,int MakeAll);
int UNREAL_API CompileScript(UClass *Class,int ActorPropertiesAreValid);

//
// Misc.
//
char *GetPropertyTypeName(EClassPropertyType Type);

/*-----------------------------------------------------------------------------
	AActorBase inlines.
-----------------------------------------------------------------------------*/

//
// See if this actor belongs to the specified class.
//
inline BOOL AActorBase::IsKindOf( const class UClass *SomeParent ) const
{
	guardSlow(AActorBase::IsKindOf);
	for( UClass *TempClass=Class; TempClass; TempClass=TempClass->ParentClass )
		if( TempClass==SomeParent )
			return 1;
	return 0;
	unguardSlow;
}

//
// See if this actor belongs to the specified class.
//
inline BOOL AActorBase::IsKindOf( const char *ClassName ) const
{
	guardSlow(AActorBase::IsKindOf);
	for( UClass *TempClass=Class; TempClass; TempClass=TempClass->ParentClass )
		if( stricmp( TempClass->Name, ClassName ) == 0 )
			return 1;
	return 0;
	unguard;
}

//
// Get the pointer to a particular actor property.
//
inline void *AActorBase::GetPropertyPtr( int iProperty,int iElement )
{
	guardSlow(AActorBase::GetPropertyPtr);
	FClassProperty &Property = Class->Element(iProperty);
	return &((AActor*)this)->Private.PropertyBins[Property.Bin]
		[
		Property.Offset + iElement * Property.ElementSize
		];
	unguardSlow;
}
inline const void *AActorBase::GetPropertyPtr( int iProperty,int iElement ) const
{
	guardSlow(AActorBase::GetPropertyPtr);
	FClassProperty &Property = Class->Element(iProperty);
	return &((AActor*)this)->Private.PropertyBins[Property.Bin]
		[
		Property.Offset + iElement * Property.ElementSize
		];
	unguardSlow;
}
inline void *AActorBase::GetPropertyPtrFromClass( UClass *Class,int iProperty,int iElement )
{
	guardSlow(AActorBase::GetPropertyPtrFromClass);
	FClassProperty &Property = Class->Element(iProperty);
	return &((AActor*)this)->Private.PropertyBins[Property.Bin]
		[
		Property.Offset + iElement * Property.ElementSize
		];
	unguardSlow;
}

// If this actor is a player, return it as a Pawn.
// Otherwise return NULL.
inline APawn *AActorBase::GetPlayer() const
{
	guardSlow(AActorBase::GetPlayer);

	// Only descendents of the pawn class may be players.
	if( !IsKindOf("Pawn") )
		return NULL;

	// Players must have cameras.
	if( ((APawn*)this)->Camera == NULL )
		return NULL;

	// This is a player.
	return (APawn*)this;

	unguardSlow;
}

// Determine if BlockingActor should block actors of the given class.
inline BOOL AActorBase::IsBlockedBy( const AActor &Other ) const
{
	guardSlow(AActorBase::IsBlockedBy);

	// Players and nonplayers are handled separately.
	if( GetPlayer () )
	{
		// Return whether the other actor blocks players.
		return bBlocksActors && Other.bBlocksPlayers;
	}
	else
	{
		// See if either actor doesn't block any other actors.
		if( !bBlocksActors || !Other.bBlocksActors )
			return 0;

		// If there are no specific blocked/passed classes everything is blocked.
		if( Other.BlockedClasses[0]==0 && Other.PassedClasses[0]==0 )
			return 1;

		// See if this class is explicitly allowed to pass.
		for( int i=0; i<3; i++ )
		{
			if( Other.PassedClasses[i] && IsKindOf(Other.PassedClasses[i]) )
				return 0;
		}

		// See if this class is explicitly blocked.
		for( i=0; i<3; i++ )
		{
			if( Other.BlockedClasses[i] && IsKindOf(Other.BlockedClasses[i]) )
				return 1;
		}

		// This actor is not blocked.
		return 0;
	}
	unguardSlow;
}

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
#endif // _INC_UNACTOR
